#include "SteamVrTrackingServer.h"




struct Config
{
	bool debug = false;
} config_;


vr::IVRSystem *vr_system_ = nullptr;
vr::TrackedDevicePose_t tracked_device_poses_[vr::k_unMaxTrackedDeviceCount];


bool initVR()
{
	
	// Loading the SteamVR Runtime
	vr::EVRInitError eError = vr::VRInitError_None;
	vr_system_ = vr::VR_Init(&eError, vr::VRApplication_Overlay);

	if (eError != vr::VRInitError_None)
	{
		vr_system_ = nullptr;
		char buf[1024];
		sprintf_s(buf, sizeof(buf), "Unable to init VR runtime: %s", vr::VR_GetVRInitErrorAsEnglishDescription(eError));
		std::cerr << buf << std::endl;
		return false;
	}

	//m_pRenderModels = (vr::IVRRenderModels *)vr::VR_GetGenericInterface(vr::IVRRenderModels_Version, &eError);
	//if (!m_pRenderModels)
	//{
	//	m_pHMD = NULL;
	//	vr::VR_Shutdown();

	//	char buf[1024];
	//	sprintf_s(buf, sizeof(buf), "Unable to get render model interface: %s", vr::VR_GetVRInitErrorAsEnglishDescription(eError));
	//	SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR, "VR_Init Failed", buf, NULL);
	//	return false;
	//}

	return true;
}

void shutdownVR()
{
	if (vr_system_)
	{
		vr::VR_Shutdown();
		vr_system_ = nullptr;
	}

	//for (std::vector< CGLRenderModel * >::iterator i = m_vecRenderModels.begin(); i != m_vecRenderModels.end(); i++)
	//{
	//	delete (*i);
	//}
	//m_vecRenderModels.clear();
}

//-----------------------------------------------------------------------------
// Purpose: Processes a single VR event
//-----------------------------------------------------------------------------
void ProcessVREvent(const vr::VREvent_t & event)
{
	switch (event.eventType)
	{
	case vr::VREvent_TrackedDeviceActivated:
	{
		printf("Device %u attached.\n", event.trackedDeviceIndex);
	}
	break;
	case vr::VREvent_TrackedDeviceDeactivated:
	{
		printf("Device %u detached.\n", event.trackedDeviceIndex);
	}
	break;
	case vr::VREvent_TrackedDeviceUpdated:
	{
		printf("Device %u updated.\n", event.trackedDeviceIndex);
	}
	break;
	}
}

void handleInput()
{
	// Process SteamVR events
	vr::VREvent_t event;
	while (vr_system_->PollNextEvent(&event, sizeof(event)))
	{
		ProcessVREvent(event);
	}
}

void printMatrix(const vr::HmdMatrix34_t &htm)
{
	for (int rows = 0; rows < 4; ++rows)
	{
		std::cout << (rows == 0 ? "[" : "") << "\t";

		if (rows == 3)
		{
			for (int cols = 0; cols < 4; ++cols)
			{
				printf("%.3f\t", (cols == 3 ? 1.0f : 0.0f));
				//std::cout << (cols == 3 ? 1 : 0) << "\t";
			}
		}
		else
		{
			for (int cols = 0; cols < 4; ++cols)
			{
				printf("%.3f\t", htm.m[rows][cols]);
				//std::cout << htm.m[rows][cols] << "\t";
			}
		}

		std::cout << (rows == 3 ? "]" : "") << std::endl;
	}
}

void printVRDevices() {
	std::cout << "List of available VR devices:" << std::endl;
	for (int nDevice = 0; nDevice < vr::k_unMaxTrackedDeviceCount; ++nDevice)
	{
		vr::ETrackedDeviceClass device_class = vr_system_->GetTrackedDeviceClass(nDevice);
		if (device_class == vr::TrackedDeviceClass_Invalid)
		{
			continue;
		}

		std::string type;
		switch (device_class)
		{
		case vr::TrackedDeviceClass_Controller:        type = "Controller"; break;
		case vr::TrackedDeviceClass_HMD:               type = "HMD"; break;
		case vr::TrackedDeviceClass_Invalid:           type = "Invalid"; break;
		case vr::TrackedDeviceClass_GenericTracker:    type = "GenericTracker"; break;
		case vr::TrackedDeviceClass_TrackingReference: type = "TrackingReference"; break;
		default:                                       type = "Unknown"; break;
		}

		const uint32_t size = 1024;
		char serial[size];
		vr_system_->GetStringTrackedDeviceProperty(nDevice, vr::ETrackedDeviceProperty::Prop_SerialNumber_String, serial, size);

		std::cout << "\tDevice " << nDevice << ":" << std::endl;
		std::cout << "\t\tType: " << type << std::endl;
		std::cout << "\t\tSerial: " << serial << std::endl;
	}
}

int main2(int argc, char *argv[])
{
	if (!initVR())
	{
		shutdownVR();
		return 1;
	}
	printVRDevices();

	handleInput();

	// Get device poses
	vr_system_->GetDeviceToAbsoluteTrackingPose(vr::TrackingUniverseRawAndUncalibrated, 0, tracked_device_poses_, vr::k_unMaxTrackedDeviceCount);

	// TODO: remove
	vr::HmdMatrix34_t raw_to_standing_htm = vr_system_->GetRawZeroPoseToStandingAbsoluteTrackingPose();
	std::cout << "RawZeroPoseToStandingAbsoluteTrackingPose:" << std::endl;
	printMatrix(raw_to_standing_htm);

	for (int nDevice = 0; nDevice < vr::k_unMaxTrackedDeviceCount; ++nDevice)
	{
		if (!tracked_device_poses_[nDevice].bPoseIsValid)
		{
			continue;
		}

		vr::ETrackedDeviceClass device_class = vr_system_->GetTrackedDeviceClass(nDevice);

		if (device_class == vr::TrackedDeviceClass_Invalid)
		{
			continue;
		}

		vr::HmdMatrix34_t htm = tracked_device_poses_[nDevice].mDeviceToAbsoluteTracking;

		std::string type;
		std::string model_type;
		switch (device_class)
		{
		case vr::TrackedDeviceClass_Controller:
			type = "Controller";
			model_type = "vive_controller";
			break;
		case vr::TrackedDeviceClass_HMD:
			type = "HMD";
			model_type = "vive_hmd";
			break;
		case vr::TrackedDeviceClass_Invalid:
			type = "Invalid";
			break;
		case vr::TrackedDeviceClass_GenericTracker:
			type = "GenericTracker";
			break;
		case vr::TrackedDeviceClass_TrackingReference:
			type = "TrackingReference";
			model_type = "vive_base_station";
			break;
		default:
			type = "Unknown";
			break;
		}

		if (config_.debug)
		{
			std::cout << "Pose of device " << nDevice << " (" << type << ")" << std::endl;
			printMatrix(htm);
		}
	}

	shutdownVR();

	return 0;
}
