#include "main.h"
#include "glut.h"
#include "SteamVRTrackingServer.h"
#include "SteamTracking.h"

#include <cmath>
#include <cstdio>

#include <Windows.h>
#include <Ole2.h>

#include <vector>
#include <iostream>
#include <fstream>

#include <sstream>
#include <glm/vec3.hpp>

#include <Kinect.h>

// We'll be using buffer objects to store the kinect point cloud
GLuint vboId;
GLuint cboId;

GLuint vboIdArrow;
GLuint cboIdArrow;

float arrowPoints[] = { 1, 0 , 0, 0, 1, 0, 0, 0, 1
};

// Intermediate Buffers
unsigned char rgbimage[colorwidth*colorheight*4];    // Stores RGB color image
ColorSpacePoint depth2rgb[width*height];             // Maps depth pixels to rgb pixels
CameraSpacePoint depth2xyz[width*height];			 // Maps depth pixels to 3d coordinates
glm::vec3 framePositions[width*height];
float frameColors[width*height * 3];

std::vector<glm::vec3> points;
std::vector<unsigned char> colors;

std::vector<glm::vec3> snapshotPoints;
std::vector<unsigned char> snapshotColors;

// Kinect Variables
IKinectSensor* sensor;             // Kinect sensor
IMultiSourceFrameReader* reader;   // Kinect data source
ICoordinateMapper* mapper;         // Converts between depth, color, and 3d coordinates

SteamTracking* steamTracking = nullptr;
glm::mat3x4* currentControlerPos = nullptr;

static bool writeFile = false;
static bool captureFrame = false;


void drawArrow() {
	//TETAEDRON BY HAND
	glBegin(GL_TRIANGLES);

	//front triangle
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
	glVertex3f(0.0f, 1.0f, 0.0f);
	glVertex3f(-1.0f, -1.0f, 0.0f);
	glVertex3f(1.0f, -1.0f, 0.0f);

	//right side triangle
	glColor4f(0.0f, 0.0f, 1.0f, 1.0f);
	glVertex3f(1.0f, -1.0f, 0.0f);
	glVertex3f(0.0f, 1.0f, 0.0f);
	glVertex3f(0.0f, -1.0f, -1.0f);

	//left side triangle
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glVertex3f(-1.0f, -1.0f, 0.0f);
	glVertex3f(0.0f, 1.0f, 0.0f);
	glVertex3f(0.0f, -1.0f, -1.0f);

	//bottom triangle
	glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
	glVertex3f(-1.0f, -1.0f, 0.0f);
	glVertex3f(1.0f, -1.0f, 0.0f);
	glVertex3f(0.0f, -1.0f, -1.0f);

	glEnd();
}

bool initKinect() {
    if (FAILED(GetDefaultKinectSensor(&sensor))) {
		return false;
	}
	if (sensor) {
		sensor->get_CoordinateMapper(&mapper);

		sensor->Open();
		sensor->OpenMultiSourceFrameReader(
			FrameSourceTypes::FrameSourceTypes_Depth | FrameSourceTypes::FrameSourceTypes_Color,
			&reader);
		return reader?true:false;
	} else {
		return false;
	}
}

void getDepthData(IMultiSourceFrame* frame) {
	IDepthFrame* depthframe;
	IDepthFrameReference* frameref = NULL;
	frame->get_DepthFrameReference(&frameref);
	frameref->AcquireFrame(&depthframe);
	if (frameref) frameref->Release();

	if (!depthframe) return;

	// Get data from frame
	unsigned int sz;
	unsigned short* buf;
	depthframe->AccessUnderlyingBuffer(&sz, &buf);

	// Write vertex coordinates
	mapper->MapDepthFrameToCameraSpace(width*height, buf, width*height, depth2xyz);

	for (unsigned int i = 0; i < sz; i++) {
		framePositions[i] = glm::vec3(depth2xyz[i].X, depth2xyz[i].Y, depth2xyz[i].Z);
	}

	// Fill in depth2rgb map
	mapper->MapDepthFrameToColorSpace(width*height, buf, width*height, depth2rgb);
	if (depthframe) depthframe->Release();
}

void getRgbData(IMultiSourceFrame* frame, bool writFrameToFile) {
	IColorFrame* colorframe;
	IColorFrameReference* frameref = NULL;
	frame->get_ColorFrameReference(&frameref);
	frameref->AcquireFrame(&colorframe);
	if (frameref) frameref->Release();

	if (!colorframe) return;

	// Get data from frame
	colorframe->CopyConvertedFrameDataToArray(colorwidth*colorheight*4, rgbimage, ColorImageFormat_Rgba);

	// Write color array for vertices
	float* fdest = (float*)frameColors;
	for (int i = 0; i < width*height; i++) {
		ColorSpacePoint p = depth2rgb[i];
		// Check if color pixel coordinates are in bounds
		if (p.X < 0 || p.Y < 0 || p.X > colorwidth || p.Y > colorheight) {
			*fdest++ = 1;
			*fdest++ = 0;
			*fdest++ = 0;
		}
		else {
			int idx = (int)p.X + colorwidth*(int)p.Y;
			*fdest++ = rgbimage[4 * idx + 0] / 255.f;
			*fdest++ = rgbimage[4 * idx + 1] / 255.f;
			*fdest++ = rgbimage[4 * idx + 2] / 255.f;

			if (writFrameToFile) {
				snapshotPoints.push_back(framePositions[i]);

				snapshotColors.push_back(rgbimage[4 * idx + 0]);
				snapshotColors.push_back(rgbimage[4 * idx + 1]);
				snapshotColors.push_back(rgbimage[4 * idx + 2]);
			}
		}
		// Don't copy alpha channel
	}

	if (colorframe) colorframe->Release();
}

void getKinectData(bool writeFrameInVector) {
	IMultiSourceFrame* frame = NULL;
	if (SUCCEEDED(reader->AcquireLatestFrame(&frame))) {
		delete currentControlerPos;
		currentControlerPos = steamTracking->getTransformationForDevice(0);

		if (writeFrameInVector) {
			snapshotPoints.clear();
			snapshotPoints.reserve(width*height);
			snapshotColors.clear();
			snapshotColors.reserve(width*height * 3);
		}

		getDepthData(frame);
		getRgbData(frame, writeFrameInVector);

		if (writeFrameInVector) {
			//get controler pos

			//frame -> global
			points.reserve(points.capacity() + width*height);
			colors.reserve(colors.capacity() + width*height);

			for(int i = 0; i < width*height; i++) {

				points.push_back(*currentControlerPos * framePositions[i]);
				colors.push_back(snapshotColors[i]);


			}

		}

		glBindBuffer(GL_ARRAY_BUFFER, vboId);
		glBufferData(GL_ARRAY_BUFFER, width*height * 3 * sizeof(GLfloat), framePositions, GL_DYNAMIC_DRAW);
		glUnmapBuffer(GL_ARRAY_BUFFER);

		glBindBuffer(GL_ARRAY_BUFFER, cboId);
		glBufferData(GL_ARRAY_BUFFER, width*height * 3 * sizeof(GLfloat), frameColors, GL_DYNAMIC_DRAW);

		glUnmapBuffer(GL_ARRAY_BUFFER);
	}
	if (frame) frame->Release();
}

void writePointCloud() {
	if (points.size() == 0) return;

	pntsHeader header;
	size_t pointArrayByteSize = points.size() * sizeof(glm::vec3);
	size_t coloArrayByteSize = colors.size() * sizeof(unsigned char);
	header.featureTableByteLenght = pointArrayByteSize + coloArrayByteSize;

	std::ostringstream os;
	os << "{\"POINTS_LENGTH\":" << (points.size() / 3) << ", \"POSITION\" : {\"byteOffset\":0}, \"RGB\" : {\"byteOffset\":" << pointArrayByteSize << "}}";
	std::string json = os.str();

	header.featureTableJSONByteLenght = json.size();
	header.byteLenght = 28 /*header*/ + header.featureTableJSONByteLenght + header.featureTableByteLenght;
	std::cout << "File Size " << header.byteLenght;

	std::ofstream myfile;
	myfile.open("example.pnts", std::ios::out | std::ios::binary);
	myfile.write((char*)&header, 28);
	myfile << json;
	myfile.write(reinterpret_cast<const char*>(&points[0]), pointArrayByteSize);
	myfile.write(reinterpret_cast<const char*>(&colors[0]), coloArrayByteSize);
	//myfile << 


	myfile.close();
}

void rotateCamera() {
	static double angle = 0.;
	static double radius = 4.;
	double x = radius*sin(angle);
	double z = radius*(1-cos(angle)) - radius/2;
	
	//gluLookAt(x,0,z,0,0,radius/2,0,1,0);
	angle += 0.002;
}

void drawKinectData() {

	getKinectData(captureFrame);
	captureFrame = false;
	if (writeFile) {
		writePointCloud();
		writeFile = false;
	}

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();


	rotateCamera();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);

	glBindBuffer(GL_ARRAY_BUFFER, vboId);
	glVertexPointer(3, GL_FLOAT, 0, NULL);

	glBindBuffer(GL_ARRAY_BUFFER, cboId);
	glColorPointer(3, GL_FLOAT, 0, NULL);

	glPointSize(1.f);
	glDrawArrays(GL_POINTS, 0, width*height);

	glPushMatrix();
	float scale = 1.0 / 10;
	glScalef(scale, scale, scale);
	
	//drawArrow();
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
	glutSolidTeapot(1);
	glPopMatrix();

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);

}

void keyPressed(unsigned char key, int x, int y)
{
	switch (key) {
	case 'a':
		captureFrame = true;
		break;
	case 's':
		writeFile = true;
		break;
	default:
		break;
	}
	glutPostRedisplay(); /* this redraws the scene without
						 waiting for the display callback so that any changes appear
						 instantly */
}


int main(int argc, char* argv[]) {
	std::cout << "Starting Kinect VR Point Cloud creator";
    if (!init(argc, argv)) return 1;
    if (!initKinect()) return 1;



	if (!initVR())
	{
		shutdownVR();
	}

	steamTracking = new SteamTracking();


    // OpenGL setup
    glClearColor(0,0,0,0);
    glClearDepth(1.0f);

	// Set up array buffers
	const int dataSize = width*height * 3 * 4;
	glGenBuffers(1, &vboId);
	glBindBuffer(GL_ARRAY_BUFFER, vboId);
	glBufferData(GL_ARRAY_BUFFER, dataSize, 0, GL_DYNAMIC_DRAW);
	glGenBuffers(1, &cboId);
	glBindBuffer(GL_ARRAY_BUFFER, cboId);
	glBufferData(GL_ARRAY_BUFFER, dataSize, 0, GL_DYNAMIC_DRAW);


    // Camera setup
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluPerspective(45, width /(GLdouble) height, 0.1, 1000);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
	gluLookAt(0,0,0,0,0,1,0,1,0);

    // Main loop
	atexit(close);
    execute();
	//drawKinectData();

	

	
    return 0;
}

void close() {
	delete steamTracking;
}
